/*
 * OOP class of RGTP connections
 * implements login and reconnect procedure
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

static int RG_BannerMeaning(Tcl_Interp *interp, RG_Connection *connP,
                            const char *s, int authdone, int wantaccess,
                            int closeifbad) {
  int oldac;
  /* wantaccess=-1 means default, as usual
   * -2 means this is initial login, and so 0 and 1 are allowed
   */
  
  if (!strncmp(s, closeifbad ? "482 " : "432 ", 4)) {
    Tcl_AppendResult(interp, "server rejected our access request:\n", s, (char*)0);
    Tcl_SetErrorCode(interp, "RGTP","BADUSER",(char*)0);
    if (closeifbad) RG_Break(connP);
    return TCL_ERROR;
  }
  if (s[0] != '2' ||
      s[1] != '3' ||
      !strchr("0123",s[2]) ||
      s[3] != ' ') {
    Tcl_AppendResult(interp,
                     "server refuses to talk to you:\n",
                     s, (char*)0);
    if (closeifbad) RG_Break(connP);
    return TCL_ERROR;
  }
  oldac= connP->access;
  connP->access= atoi(s+2);
  if ((wantaccess >=0 && connP->access != wantaccess) ||
      (wantaccess ==-2 && connP->access > 1)) {
    char buf1[10], buf2[10];
    sprintf(buf1,"%d",wantaccess);
    sprintf(buf2,"%d",connP->access);
    Tcl_AppendResult(interp,
                     "protocol error by server - unexpected access level ",buf2,
                     " (expected ",buf1,"):\n", s, (char*)0);
    if (closeifbad) RG_Violation(connP);
    return TCL_ERROR;
  }
  if (!authdone && wantaccess != -2 && connP->md5secretbytes && connP->access > oldac) {
    Tcl_AppendResult(interp,
                     "POSSIBLE BOGUS SERVER DETECTED: We were expecting to do mutual"
                     " authentication, but the server didn't want to.  Please"
                     " contact the server administrators at once.\n", s, (char*)0);
    Tcl_SetErrorCode(interp, "RGTP","BADSERVER","NOAUTH",(char*)0);
    if (closeifbad) RG_Break(connP);
    return TCL_ERROR;
  }

  return TCL_OK;
}

static int matchhexpattern(const char *mp, char *p) {
  while (*mp && (*mp=='.' ?
                 strchr("0123456789ABCDEF",toupper(*p))!=0 :
                 *p==*mp)) { p++; mp++; }
  return !*mp && !*p;
}

int RG_HandleAuth(Tcl_Interp *interp, RG_Connection *connP,
                  Tcl_DString *bannerP, int wantaccess, int closeifbad) {
  /* Will call Tcl_DStringFree on bannerP */
  int authdone;
  int r;
  Tcl_DString ds;

  authdone= 0;
  if (!strncmp(Tcl_DStringValue(bannerP),"130 MD5 ",8)) {
    unsigned char clientnonce[16];
    unsigned char servernonce[16]; 
    unsigned char messagebuf[16*3+SECRET_MAXBYTES];
    MD5_CTX md5ctx;
    struct sockaddr_in saddr;
    struct timeval timevab;
    int saddrl, i, v;
    unsigned short us;
    char *p, *q, cbuf[100];

    if (connP->md5secretbytes <= 0) {
      Tcl_AppendResult(interp,
                       "Server requested authentication details"
                       " which I don't have:\n",
                       Tcl_DStringValue(bannerP), (char*)0);
      Tcl_DStringFree(bannerP);
      RG_Break(connP);
      Tcl_SetErrorCode(interp, "RGTP","BADSECRET","MISSING",(char*)0);
      return TCL_ERROR;
    }
    
    saddrl= sizeof(saddr);
    r= getsockname(fileno(connP->file),(struct sockaddr*)&saddr,&saddrl);
    if (r) {
      p= Tcl_PosixError(interp);
      Tcl_AppendResult(interp, "couldn't find own network address: ",
                       p, (char*)0);
      RG_Break(connP); return TCL_ERROR;
    }
    if (saddrl != sizeof(saddr)) {
      sprintf(interp->result,
              "own network address length wrong (%d, expected %d)",
              saddrl, (int)sizeof(saddr));
      RG_Break(connP); return TCL_ERROR;
    }
    if (gettimeofday(&timevab,(void*)0)) {
      p= Tcl_PosixError(interp);
      Tcl_AppendResult(interp, "couldn't gettimeofday for nonce: ",
                       p, (char*)0);
      RG_Break(connP); return TCL_ERROR;
    }
    memcpy(clientnonce,&saddr.sin_addr,4);
    memcpy(clientnonce+4,&timevab.tv_sec,4);
    memcpy(clientnonce+8,&timevab.tv_usec,4);
    memcpy(clientnonce+12,&saddr.sin_port,2);
    us= getpid(); memcpy(clientnonce+14,&us,2);

    Tcl_DStringFree(bannerP);
    r= RG_ReadLineEcho(interp, connP, &ds); if (r != TCL_OK) return r;
    if (!matchhexpattern("333 ................................",
                         Tcl_DStringValue(&ds))) {
      Tcl_AppendResult(interp,
                       "server challenge after 130 MD5 malformatted:\n",
                       Tcl_DStringValue(&ds), (char*)0);
      Tcl_DStringFree(&ds); RG_Break(connP); return TCL_ERROR;
    }
    for (i=0; i<16; i++) {
      sscanf(Tcl_DStringValue(&ds)+4+i*2,"%2x",&v); servernonce[i]= v;
    }
    
    memcpy(messagebuf,clientnonce,16);
    memcpy(messagebuf+16,servernonce,16);
    memset(messagebuf+32,0,16); strncpy(messagebuf+32,connP->user,16);
    for (p=connP->md5secret, q=messagebuf+48, i=connP->md5secretbytes;
         i>0;
         i--) *q++= ~*p++;
    MD5Init(&md5ctx);
    if (connP->debug) {
      fprintf(stderr,"RGTP %d (md5 ",connP->id);
      for (i=0; i<48; i++) fprintf(stderr,"%02X",messagebuf[i]);
      fprintf(stderr,"!!%d)\n",connP->md5secretbytes);
    }
    MD5Update(&md5ctx,messagebuf,48+connP->md5secretbytes);
    MD5Final(&md5ctx);

    for (i=0; i<16; i++) sprintf(cbuf+i*2,"%02X",md5ctx.digest[i]);
    cbuf[32]= ' ';
    for (i=0; i<16; i++) sprintf(cbuf+33+i*2,"%02X",clientnonce[i]);
    r= RG_CommandResponse(interp, connP, &ds, "AUTH %s", cbuf);
    if (r != TCL_OK) return r;
    if (!strncmp(Tcl_DStringValue(&ds),closeifbad ? "483 " : "433 ",4)) {
      Tcl_SetErrorCode(interp, "RGTP","BADSECRET","INCORRECT",(char*)0);
      Tcl_AppendResult(interp,
                       "invalid secret or password:\n",
                       Tcl_DStringValue(&ds), (char*)0);
      if (closeifbad) RG_Break(connP);
      Tcl_DStringFree(&ds);
      return TCL_ERROR;
    } else if (!matchhexpattern("133 ................................",
                                Tcl_DStringValue(&ds))) {
      Tcl_AppendResult(interp,
                       "server response to our AUTH malformatted:\n",
                       Tcl_DStringValue(&ds), (char*)0);
      Tcl_DStringFree(&ds); RG_Break(connP); return TCL_ERROR;
    }

    memcpy(messagebuf,servernonce,16);
    memcpy(messagebuf+16,clientnonce,16);
    memset(messagebuf+32,0,16); strncpy(messagebuf+32,connP->user,16);
    memcpy(messagebuf+48,connP->md5secret,connP->md5secretbytes);
    MD5Init(&md5ctx);
    if (connP->debug) {
      fprintf(stderr,"RGTP %d (md5 ",connP->id);
      for (i=0; i<48; i++) fprintf(stderr,"%02X",messagebuf[i]);
      fprintf(stderr,"??%d)\n",connP->md5secretbytes);
    }
    MD5Update(&md5ctx,messagebuf,48+connP->md5secretbytes);
    MD5Final(&md5ctx);
    if (connP->debug) {
      fprintf(stderr,"RGTP %d (md5=",connP->id);
      for (i=0; i<16; i++) fprintf(stderr,"%02X",md5ctx.digest[i]);
      fprintf(stderr,")\n");
    }
    for (i=0; i<16; i++) {
      sscanf(Tcl_DStringValue(&ds)+4+i*2,"%2x",&v);
      if (md5ctx.digest[i] != v) {
        Tcl_AppendResult(interp,
                         "BOGUS SERVER DETECTED - please"
                         " contact the server administrators at once!",
                         (char*)0);
        Tcl_SetErrorCode(interp, "RGTP","BADSERVER","BADAUTH",(char*)0);
        Tcl_DStringFree(&ds); RG_Break(connP); return TCL_ERROR;
      }
    }
    Tcl_DStringFree(&ds);
    bannerP= &ds;
    r= RG_ReadLineEcho(interp, connP, bannerP); if (r != TCL_OK) return r;
    authdone= 1;
  }
  r= RG_BannerMeaning(interp, connP, Tcl_DStringValue(bannerP), authdone,
                      wantaccess, closeifbad);
  Tcl_DStringFree(bannerP);
  if (r != TCL_OK) return r;
  if (connP->wantaccess == -1) connP->wantaccess= connP->access;
  return TCL_OK;
}

int RG_Login(Tcl_Interp *interp, RG_Connection *connP) {
  int r;
  Tcl_DString ds;
  char buf[10];

  Tcl_ResetResult(interp);
  buf[0]=0; if (connP->wantaccess != -1) sprintf(buf," %d",connP->wantaccess);
    
  r= RG_CommandResponseRetry(interp, connP, &ds, "USER %s%s", connP->user, buf);
  if (r != TCL_OK) return r;
  return RG_HandleAuth(interp,connP,&ds,connP->wantaccess,1);
}

int RG_Reconnect(Tcl_Interp *interp, RG_Connection *connP) {
  /* Status must be Broken or NeedReconnect before,
   * will be correct afterwards.
   * Result will be erased even if successful
   */
  unsigned long hostaddr, *hostaddrps[2], **hostaddrpc;
  struct hostent *hose, hoss;
  struct sockaddr_in saddr;
  char *msg;
  int r, fd, ri;
  Tcl_DString ds;

  connP->status= RG_Status_Broken;
  connP->contitem[0]= connP->editing= connP->editlocked= connP->logindone= 0;
  Tcl_ResetResult(interp);
  hostaddr= inet_addr(connP->host);
  if (hostaddr == (unsigned long)-1) {
    hose= gethostbyname(connP->host);
    if (!hose) {
      Tcl_AppendResult(interp, "unknown host \"",
                       connP->host, "\"", (char*)0);
      return TCL_ERROR;
    }
    if (hose->h_length != sizeof(hostaddr) ||
        hose->h_addrtype != AF_INET) {
      sprintf(interp->result,
              "unknown address length (%ld) or type (%ld) "
              "returned by gethostbyname",
              (long)hose->h_length, (long)hose->h_addrtype);
      return TCL_ERROR;
    }
  } else {
    hostaddrps[0]= &hostaddr;
    hostaddrps[1]= 0;
    hose= &hoss;
    hoss.h_addr_list= (char**)hostaddrps;
  }
  fd= socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (fd < 0) {
    msg= Tcl_PosixError(interp); 
    Tcl_AppendResult(interp, "failed to create socket: ", msg, (char*)0);
    return TCL_ERROR;
  }
  interp->result= "failed to connect to server";
  r= 1; /* none tried yet */
  for (hostaddrpc= (unsigned long **)hose->h_addr_list;
       *hostaddrpc;
       hostaddrpc++) {
    saddr.sin_family = AF_INET;
    saddr.sin_port= connP->port;
    *(unsigned long*)&saddr.sin_addr= **hostaddrpc;
    memset((char*)&saddr.sin_zero, 0, sizeof(saddr.sin_zero));
    ri= RG_StartIntIO(interp,fd);  if (ri != TCL_OK) return ri;
    do {
      if (connP->debug) fprintf(stderr,"RGTP %d (connect %s,%d)\n",
                                connP->id, inet_ntoa(saddr.sin_addr),
                                ntohs(connP->port));
      r= connect(fd, (struct sockaddr*)&saddr, sizeof(saddr));
    } while (r && errno == EINTR);
    ri= RG_EndIntIO(interp,0);  if (ri != TCL_OK) return ri;
    if (!r) break;
    msg= Tcl_PosixError(interp);
    Tcl_AppendResult(interp,
                     "\n(tried ", inet_ntoa(saddr.sin_addr),
                     ": ", msg, ")", (char*)0);
  }
  if (r) {
    if (r >= 1) {
      Tcl_AppendResult(interp, " (host \"", connP->host,
                       "\" exists but has no addresses)", (char*)0);
    }
    return TCL_ERROR;
  }

  Tcl_ResetResult(interp);
  connP->file= fdopen(fd, "r+");
  if (!connP->file) {
    msg= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't turn socket into stdio stream: ",
                     msg, (char*)0);
    close(fd);
    return TCL_ERROR;
  }
  if (setvbuf(connP->file, (char*)0, _IOFBF, BUFSIZ)) {
    msg= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't setvbuf on socket streams: ",
                     msg, (char*)0);
    fclose(connP->file); close(fd);
    return TCL_ERROR;
  }
  if (connP->debug) fprintf(stderr,"RGTP %d (connected)\n", connP->id);
  r= RG_ReadLineEcho(interp, connP, &ds);
  if (r != TCL_OK) return r;
  r= RG_BannerMeaning(interp, connP, Tcl_DStringValue(&ds), 0,-2,1);
  Tcl_DStringFree(&ds);
  if (r != TCL_OK) return r;
  connP->status= RG_Status_OK;
  if (!connP->user || connP->access == connP->wantaccess) return TCL_OK;
  else return RG_Login(interp, connP);
}
