/*
 * OOP class of RGTP connections
 * assorted utility, communication and management functions
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

void RG_EArrayInit(RG_EArray *eaP) {
  eaP->data= 0;
  eaP->allocd= eaP->used= 0;
}

void RG_EArrayFree(RG_EArray *eaP) {
  if (eaP->data) { ckfree(eaP->data); }
  RG_EArrayInit(eaP);
}

void RG_EArrayReset(RG_EArray *eaP) {
  eaP->used= 0;
}

void *RG_EArrayAppend(RG_EArray *eaP, size_t sz) {
  if (eaP->used >= eaP->allocd) {
    eaP->allocd += 10;
    eaP->allocd <<= 1;
    eaP->data= eaP->data ? ckrealloc(eaP->data, sz * eaP->allocd)
                         : ckalloc(sz * eaP->allocd);
  }
  return (void*)((char*)eaP->data + (sz * eaP->used++));
}

void RG_EArrayEnsureSize(RG_EArray *eaP, int n, size_t sz) {
  if (eaP->allocd >= n) return;
  eaP->data= eaP->data ? ckrealloc(eaP->data, sz*n)
                       : ckalloc(sz*n);
  eaP->allocd= n;
}

void RG_Break(RG_Connection *connP) {
  if (connP->debug) fprintf(stderr,"RGTP %d (break)\n",connP->id);
  close(fileno(connP->file)); fclose(connP->file);
  connP->status= RG_Status_Broken;
  connP->editing= connP->editlocked= connP->regudone= connP->logindone= 0;
  connP->contitem[0]= 0;
}

void RG_Violation(RG_Connection *connP) {
  if (connP->debug) fprintf(stderr,"RGTP %d (violation)\n",connP->id);
  close(fileno(connP->file)); fclose(connP->file);
  connP->status= RG_Status_NeedReconnect;
  connP->editing= connP->editlocked= connP->logindone= 0;
  connP->contitem[0]= 0;
}

static const int intiosignals[]= { SIGHUP, SIGINT, SIGPIPE, 0 };
static volatile int intioclosefd= -1;
static struct sigaction intiosigacts[sizeof(intiosignals)/sizeof(int)];

static void intiosighandler(int dummy) {
  int e, f;
  e= errno; f= intioclosefd;
  if (f != -1) close(f);
  intioclosefd= -1; errno= e;
}

int RG_StartIntIO(Tcl_Interp *interp, int fd) {
  struct sigaction nsa;
  int i,r;
  sigset_t nullset;
  const int *ip;

  assert(intioclosefd == -1);
  intioclosefd= fd;
  sigemptyset(&nullset);
  for (i=0, ip= intiosignals; *ip; i++, ip++) {
    nsa.sa_handler= intiosighandler;
    nsa.sa_mask= nullset;
    nsa.sa_flags= SA_INTERRUPT;
    r= sigaction(*ip,&nsa,&intiosigacts[i]);
    if (r) { perror("fatal error: sigaction failed in RG_StartIntIO"); _exit(1); }
  }
  return TCL_OK;
}

int RG_EndIntIO(Tcl_Interp *interp, RG_Connection *connP) {
  int i,r;
  const int *ip;
  for (i=0, ip= intiosignals; *ip; i++, ip++) {
    r= sigaction(*ip,&intiosigacts[i],0);
    if (r) { perror("fatal error: sigaction failed in RG_EndIntIO"); _exit(1); }
  }
  if (intioclosefd == -1) {
    Tcl_SetErrorCode(interp, "RGTP","INT",(char*)0);
    interp->result= "interrupted during conversation with server.";
    if (connP) RG_Break(connP);
    return TCL_ERROR;
  }
  intioclosefd= -1;
  return TCL_OK;
}

int RG_Flush(Tcl_Interp *interp, RG_Connection *connP) {
  int ri, e, r;
  char *p;
  
  ri= RG_StartIntIO(interp,fileno(connP->file));  if (ri != TCL_OK) return ri;
  r= fflush(connP->file);
  e= errno;  ri= RG_EndIntIO(interp,connP);  if (ri != TCL_OK) return ri;
  if (r) {
    errno= e; p= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't flush command/data server: ", p, (char*)0);
    RG_Break(connP);
    return TCL_ERROR;
  }
  fseek(connP->file,0,SEEK_CUR);
  return TCL_OK;
}

void RG_DeleteCallback(ClientData clientData) {
  RG_Connection *connP= (RG_Connection*)clientData;
  if (connP->status == RG_Status_OK) RG_Break(connP);
  if (connP->debug) fprintf(stderr,"RGTP %d (deleted)\n",connP->id);
  RG_EArrayFree(&connP->marks);
  RG_EArrayFree(&connP->indexdata);
  RG_EArrayFree(&connP->indexorder);
  Tcl_DStringFree(&connP->itemtext);
  ckfree((char*)connP->host);
  ckfree((char*)connP->user);
  ckfree((char*)connP);
}

int RG_ReadLine(Tcl_Interp *interp,
                RG_Connection *connP,
                Tcl_DString *dsP) {
  char buf[100], *p;
  int l, e, ri;

  Tcl_DStringInit(dsP);
  do {
    ri= RG_StartIntIO(interp,fileno(connP->file));  if (ri != TCL_OK) return ri;
    p= fgets(buf,sizeof(buf),connP->file); e= errno;
    ri= RG_EndIntIO(interp,connP);  if (ri != TCL_OK) return ri;
    if (!p) {
      if (ferror(connP->file)) {
        errno= e; p= Tcl_PosixError(interp);
      } else {
        p= "server closed connection";
      }
      if (connP->debug) fprintf(stderr,"RGTP %d (read) %s\n", connP->id, p);
      Tcl_AppendResult(interp, "error reading from server: ",
                       p, (char*)0);
      RG_Break(connP);
      return TCL_ERROR;
    }
    l= strlen(p);
    if (buf[l-1] == '\n') { l--; if (buf[l-1] == '\r') l--; }
    Tcl_DStringAppend(dsP,buf,l);
  } while (!buf[l]);
  return TCL_OK;
}

int RG_ReadLineEcho(Tcl_Interp *interp, RG_Connection *connP,
                Tcl_DString *dsP) {
  int r;
  r= RG_ReadLine(interp, connP, dsP);
  if (r != TCL_OK) return r;
  if (connP->debug) fprintf(stderr,"RGTP %d << %s\n", connP->id, dsP->string);
  return TCL_OK;
}

int RG_CommandResponse(Tcl_Interp *interp, RG_Connection *connP,
                       Tcl_DString *dsP,
                       char *format, ...) {
  va_list va;
  char *p;
  int r, ri;

  if (connP->status == RG_Status_NeedReconnect) {
    r= RG_Reconnect(interp, connP);  if (r != TCL_OK) return r;
  }
  r= RG_Flush(interp,connP); if (r != TCL_OK) return r;
  ri= RG_StartIntIO(interp,fileno(connP->file));  if (ri != TCL_OK) return ri;
  va_start(va, format);
  r= vfprintf(connP->file, format, va);
  va_end(va);
  if (r >= 0) r= fputs("\r\n",connP->file) == EOF ? -1 : 0;
  ri= RG_EndIntIO(interp,connP);  if (ri != TCL_OK) return ri;
  if (r < 0) {
    p= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't send command to server: ", p, (char*)0);
    RG_Break(connP);
    return TCL_ERROR;
  }
  r= RG_Flush(interp,connP); if (r != TCL_OK) return r;
  if (connP->debug) {
    fprintf(stderr,"RGTP %d -> ", connP->id);
    va_start(va, format);
    vfprintf(stderr,format,va);
    va_end(va);
    fprintf(stderr,"\n");
  }
  r= RG_ReadLineEcho(interp, connP, dsP);
  if (r == TCL_OK && dsP->string[0] == '5') {
    Tcl_AppendResult(interp,
                     "protocol error by client or server (this is a bug); "
                     "server said:\n", dsP->string, (char*)0);
    Tcl_DStringFree(dsP);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  return r;
}

int RG_CommandResponseRetry(Tcl_Interp *interp, RG_Connection *connP,
                            Tcl_DString *dsP,
                            char *format, ...) {
  va_list va;
  int r, ri, attempts=3;
  char *p;

  if (connP->status == RG_Status_NeedReconnect) {
    r= RG_Reconnect(interp, connP);  if (r != TCL_OK) return r;
  }
  while (attempts--) {
    r= RG_Flush(interp,connP); if (r != TCL_OK) return r;
    ri= RG_StartIntIO(interp,fileno(connP->file));  if (ri != TCL_OK) return ri;
    va_start(va, format);
    r= vfprintf(connP->file, format, va);
    va_end(va);
    if (r != EOF) r= fputs("\r\n",connP->file);
    ri= RG_EndIntIO(interp,connP);  if (ri != TCL_OK) return ri;
    if (r == EOF) {
      p= Tcl_PosixError(interp);
      Tcl_AppendResult(interp, "couldn't send command to server: ", p, (char*)0);
      RG_Break(connP);
      return TCL_ERROR;
    }
    r= RG_Flush(interp,connP); if (r != TCL_OK) return r;
    if (connP->debug) {
      fprintf(stderr,"RGTP %d => ", connP->id);
      va_start(va, format);
      vfprintf(stderr,format,va);
      va_end(va);
      fprintf(stderr,"\n");
    }
    r= RG_ReadLineEcho(interp, connP, dsP);
    if (r == TCL_OK && dsP->string[0] == '5') {
      Tcl_AppendResult(interp,
                       "protocol error by client or server (this is a bug); "
                       "server said:\n", dsP->string, (char*)0);
      Tcl_DStringFree(dsP);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    if (r != TCL_OK || strncmp(dsP->string,"481 ",4)) return r;
    Tcl_DStringFree(dsP);
    
    /* inactivity timeout */
    RG_Break(connP);
    r= RG_Reconnect(interp, connP);
    if (r != TCL_OK) return r;
  }
  Tcl_SetResult(interp,
                "repeated timeouts - network slow or server broken",
                TCL_STATIC);
  return TCL_ERROR;
}

int RG_TransmitData(Tcl_Interp *interp, RG_Connection *connP,
                    char *grogname, char *text, const RG_KeyValuePair *errortable) {
  int r, ri;
  Tcl_DString ds;
  char *p, *q;
  long l;
  
  r= RG_CommandResponseRetry(interp,connP,&ds, "DATA");
  if (r != TCL_OK) return r;
  if (strncmp(Tcl_DStringValue(&ds),"150 ",4)) {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp,
                     "server sent bad response to DATA:\n",
                     Tcl_DStringValue(&ds), (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  
  r= RG_Flush(interp,connP); if (r != TCL_OK) return r;
  ri= RG_StartIntIO(interp,fileno(connP->file));  if (ri != TCL_OK) return ri;
  r= 0;
  if (grogname) {
    if (strlen(grogname) > TXRXLINE_MAXLEN || strchr(grogname,'\n')) {
      RG_EndIntIO(interp,connP);  if (ri != TCL_OK) return ri;
      interp->result= "grossly malformatted grogname"; return TCL_ERROR;
    }
    if (fputs(grogname,connP->file) == EOF) r= -1;
    if (!r && fputs("\r\n",connP->file) == EOF) r= -1;
  }
  for (p= text;  !r && p && *p;  p= q ? q+1 : 0) {
    q= strchr(p,'\n');
    l= q ? (long)(q-p) : strlen(p);
    if (l > TXRXLINE_MAXLEN) break;
    if (*p == '.') if (fputc('.',connP->file) == EOF) r= -1;
    if (fwrite(p,1,(int)l,connP->file) != l) r= -1;
    if (!r && fputs("\r\n",connP->file) == EOF) r= -1;
  }
  if (!r && fputs(".\r\n",connP->file) == EOF) r= -1;
  ri= RG_EndIntIO(interp,connP);  if (ri != TCL_OK) return ri;
  if (r) {
    p= Tcl_PosixError(interp);
    Tcl_AppendResult(interp, "couldn't send data to server: ",
                     p, (char*)0);
    RG_Break(connP);
    return TCL_ERROR;
  }
  if (p && *p) {
    interp->result= "FORMAT TEXT {Text for reply has line that is far too long.}";
    return TCL_RETURN;
  }
  r= RG_Flush(interp,connP); if (r != TCL_OK) return r;
  r= RG_ReadLineEcho(interp, connP, &ds);
  if (r != TCL_OK) return r;
  if (RG_SubmitError(interp,&ds,errortable)) return TCL_RETURN;
  q= Tcl_DStringValue(&ds);
  if (strncmp(q,"350 ",4)) {
    Tcl_AppendResult(interp,
                     "server sent bad response after data sent:\n",
                     q, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  return TCL_OK;
}

int RG_ReadData(Tcl_Interp *interp, RG_Connection *connP,
                Tcl_DString *rdsP) {
  Tcl_DString lds;
  Tcl_DString llds;
  char *p= 0;
  int r;
  
  Tcl_DStringInit(rdsP);
  Tcl_DStringInit(&llds);
  for (;;) {
    r= RG_ReadLine(interp, connP, &lds);
    if (r != TCL_OK) {
      Tcl_AppendResult(interp, Tcl_DStringValue(&llds), (char*)0);
    } else {
      p= Tcl_DStringValue(&lds);
      if (*p == '.') {
        ++p;
        if (!*p) {
          if (connP->debug) fprintf(stderr,"RGTP %d << .\n", connP->id);
          Tcl_DStringFree(&lds);
          Tcl_DStringFree(&llds);
          return TCL_OK;
        } else if (*p != '.') {
          RG_Violation(connP);
          Tcl_AppendResult(interp,
                           "server has sent brokenly dot-stuffed line, "
                           "closing connection:\n",
                           Tcl_DStringValue(&lds), (char*)0);
          r= TCL_ERROR;
          Tcl_DStringFree(&lds);
        }
      }
    }
    if (r != TCL_OK) {
      Tcl_DStringFree(&llds);
      Tcl_DStringFree(rdsP);
      return r;
    }
    Tcl_DStringAppend(rdsP,p,-1);
    Tcl_DStringAppend(rdsP,"\n",1);
    Tcl_DStringFree(&llds);
    Tcl_DStringAppend(&llds,"\nlast line received from server was:\n",-1);
    Tcl_DStringAppend(&llds,Tcl_DStringValue(&lds),-1);
    Tcl_DStringFree(&lds);
  }
}

char *RG_CommandName(RG_Connection *connP) {
  static char buf[50];
  sprintf(buf, "rgtp-connection%d", connP->id);
  return buf;
}

int RG_GetAccessLevel(Tcl_Interp *interp, char *s, int *rhere) {
  static char *table[]= { "read", "write", "edit", 0 };
  char **tp;
  int v, r;
  
  for (tp=table, v=1; *tp && strcmp(*tp,s); tp++, v++);
  if (*tp) { *rhere=v; return TCL_OK; }
  r= Tcl_GetInt(interp,s,&v);
  if (r != TCL_OK) {
    Tcl_AppendResult(interp,"; alternatively would accept access level word",(char*)0);
    return r;
  }
  if (r >= 1 || r <= 3) { *rhere=v; return TCL_OK; }
  interp->result= "access level integer out of range (1..3)";
  return TCL_ERROR;
}

static int scanhex(const char *p, unsigned char *dest, int maxn) {
  static const char hexdigits[]= "0123456789ABCDEF";
  int c, v, n= 0;
  char *e;

  while (*p) {
    if (++n > maxn) return -1;
    c= *p++; e= strchr(hexdigits,toupper(c)); if (!e) return -1;
    v= (int)(e-hexdigits) << 4;
    c= *p++; e= strchr(hexdigits,toupper(c)); if (!e) return -1;
    *dest++ = v | (int)(e-hexdigits);
  }
  return n;
}

int RG_GetMD5Secret(Tcl_Interp *interp, RG_Connection *connP, char *s) {
  if (*s) {
    connP->md5secretbytes= scanhex(s,connP->md5secret,SECRET_MAXBYTES);
    if (connP->md5secretbytes <= 0) {
      Tcl_AppendResult(interp,
                       "badly formatted MD5 secret \"", s,
                       "\": should be pairs of hex digits or empty",
                       (char*)0);
      Tcl_SetErrorCode(interp, "RGTP","BADSECRET","SYNTAX",(char*)0);
      return TCL_ERROR;
    }
  } else {
    connP->md5secretbytes= 0;
  }
  return TCL_OK;
}

int RG_AccessOK(Tcl_Interp *interp, RG_Connection *connP, int level) {
  int r;

  if (connP->status == RG_Status_NeedReconnect) {
    r= RG_Reconnect(interp,connP);
    if (r != TCL_OK) return r;
  }
  switch (connP->status) {
  case RG_Status_OK:
    if (connP->access >= level) return TCL_OK;
    sprintf(interp->result,
            "rgtp: insufficient access level (have %d, need %d)",
            connP->access, level);
    return TCL_ERROR;
  case RG_Status_Broken:
    sprintf(interp->result,"rgtp: %s is broken",RG_CommandName(connP));
    return TCL_ERROR;
  default:
    abort();
  }
}

void RG_DeleteCommand(Tcl_Interp *interp, RG_Connection *connP) {
  Tcl_DeleteCommand(interp, RG_CommandName(connP));
}
